from __future__ import annotations

import sys
from ctypes import (
    cdll,
    CDLL,
    byref,
    c_int,
    c_short,
    c_char_p,
    c_ushort,
    c_wchar_p,
    create_string_buffer,
    sizeof,
    c_wchar,
)
from dataclasses import dataclass
from pathlib import Path
from subprocess import getstatusoutput  # nosec
from typing import Union, TYPE_CHECKING, Tuple


from ._cursor import Cursor

if TYPE_CHECKING:
    from ._environment import Environment
    from ._connection import Connection
from ._errors import ProgrammingError
from ._handler import Handler
from ._enums import (
    ReturnCode,
    HandleType,
    OdbcVersion,
    EnvironmentAttributeType,
    DriverCompletion,
    SqlFetchType,
    DataType,
)
from ._typedef import SQLSMALLINT, SQLLEN, SQLULEN


def find_library() -> Path:
    paths = (
        Path("/usr/lib64"),
        Path("/usr/lib"),
        Path("/usr/lib/i386-linux-gnu"),
        Path("/usr/lib/x86_64-linux-gnu"),
        Path("/usr/lib/libiodbc.dylib"),
    )
    for path in paths:
        libs = sorted(path.glob("libodbc.so*"))
        if libs:
            return libs[-1]
    raise FileNotFoundError


__lib: CDLL = cdll.LoadLibrary(str(find_library()))


def get_sqlwchar_size() -> int:
    """unixODBC defaults to 2-bytes SQLWCHAR, unless "-DSQL_WCHART_CONVERT" was
    added to CFLAGS, in which case it will be the size of wchar_t.
    Note that using 4-bytes SQLWCHAR will break most ODBC drivers, as driver
    development mostly targets the Windows platform.
    """
    status, output = getstatusoutput("odbc_config --cflags")
    if status == 0 and "SQL_WCHART_CONVERT" in output:
        return sizeof(c_wchar)
    return sizeof(c_ushort)


def get_unicode_size() -> int:
    """Determine the size of Py_UNICODE
    sys.maxunicode > 65536 and 'UCS4' or 'UCS2'
    """
    return sys.maxunicode > 65536 and 4 or 2


SQLWCHAR_SIZE = get_sqlwchar_size()
UNICODE_SIZE = get_unicode_size()


wchar_pointer = c_wchar_p
create_buffer = create_string_buffer
odbc_encoding = "utf_16_le"
odbc_decoding = "utf_16"
ucs_length = 2


def UTF16_BE_dec(buffer):
    i = 0
    uchars = []
    while True:
        # TODO: verify that this condition correctly identifies
        # a surrogate pair in UTF-16 BE
        if ord(buffer.raw[i + 1]) & 0xD0 == 0xD0:
            n = 2
        else:
            n = 1
        uchar = buffer.raw[i : i + n * ucs_length].decode(odbc_decoding)
        if uchar == str("\x00"):
            break
        uchars.append(uchar)
        i += n * ucs_length
    return "".join(uchars)


def UCS_dec(buffer):
    i = 0
    uchars = []
    while True:
        uchar = buffer.raw[i : i + ucs_length].decode(odbc_decoding)
        if uchar == str("\x00"):
            break
        uchars.append(uchar)
        i += ucs_length
    return "".join(uchars)


# This is the common case on Linux, which uses wide Python build together with
# the default unixODBC without the "-DSQL_WCHART_CONVERT" CFLAGS.
if sys.platform not in ("win32", "cli", "cygwin"):
    if UNICODE_SIZE >= SQLWCHAR_SIZE:
        # We can only use unicode buffer if the size of wchar_t (UNICODE_SIZE) is
        # the same as the size expected by the driver manager (SQLWCHAR_SIZE).
        create_buffer_u = create_buffer
        wchar_pointer = c_char_p

        def UCS_buf(s):
            return s.encode(odbc_encoding)

        if odbc_encoding == "utf_16":
            from_buffer_u = UTF16_BE_dec
        else:
            from_buffer_u = UCS_dec

    # Esoteric case, don't really care.
    # elif UNICODE_SIZE < SQLWCHAR_SIZE:
    #     raise OdbcLibraryError(
    #         "Using narrow Python build with ODBC library expecting wide unicode is not supported."
    #     )


def _handle_error(handler) -> None:
    ansi = True
    if ansi:
        state = create_buffer(22)
        message = create_buffer(1024 * 4)
        f = __lib.SQLGetDiagRec
        # raw_s = lambda s: bytes(s, "ascii")
    else:
        state = create_buffer_u(24)
        message = create_buffer_u(1024 * 4)
        f = __lib.SQLGetDiagRecW
        # raw_s = str
    native_error = c_int()
    buffer_len = c_short()
    err_list = []

    while True:
        err_number = len(err_list) + 1
        return_code = ReturnCode(
            f(
                handler.handle_type.value,
                handler.handle,
                err_number,
                state,
                byref(native_error),
                message,
                1024,
                byref(buffer_len),
            )
        )
        if return_code is ReturnCode.SQL_SUCCESS:
            if ansi:
                err_list.append((state.value, message.value, native_error.value))
            else:
                err_list.append(
                    (from_buffer_u(state), from_buffer_u(message), native_error.value)
                )
        elif return_code is ReturnCode.SQL_INVALID_HANDLE:
            raise ProgrammingError("", return_code.name)
        elif return_code == ReturnCode.SQL_NO_DATA:
            raise Exception(f"{state.value.decode()}{message.value.decode()}")
        else:
            raise Exception(f"Unhandled return code: {return_code}")


def check_success(return_code: Union[int, ReturnCode], handler: Handler) -> None:
    if isinstance(return_code, int):
        return_code = ReturnCode(return_code)

    success_codes = (
        ReturnCode.SQL_SUCCESS,
        ReturnCode.SQL_SUCCESS_WITH_INFO,
        ReturnCode.SQL_NO_DATA,
    )

    if return_code not in success_codes:
        _handle_error(handler)


ReturnCodeAndHandler = Tuple[int, Handler]


def allocate_environment(environment: Environment) -> None:
    return_code = __lib.SQLAllocHandle(
        HandleType.SQL_HANDLE_ENV.value,
        HandleType.SQL_NULL_HANDLE.value,
        byref(environment.handle),
    )
    check_success(return_code, environment)


def set_environment_odbc_version(
    environment: Environment, odbc_version: OdbcVersion
) -> None:
    attribute = EnvironmentAttributeType.SQL_ATTR_ODBC_VERSION
    return_code = __lib.SQLSetEnvAttr(
        environment.handle, attribute.value, odbc_version.value
    )
    check_success(return_code, environment)


def allocate_connection(environment: Environment, connection: Connection) -> None:
    return_code = __lib.SQLAllocHandle(
        HandleType.SQL_HANDLE_DBC.value, environment.handle, byref(connection.handle)
    )
    check_success(return_code, connection)


def allocate_statement(cursor: Cursor) -> None:
    return_code = __lib.SQLAllocHandle(
        HandleType.SQL_HANDLE_STMT.value, cursor.connection.handle, byref(cursor.handle)
    )
    check_success(return_code, cursor)


def sql_free_handle(handler: Handler) -> None:
    return_code = __lib.SQLFreeHandle(handler.handle_type.value, handler.handle)
    check_success(return_code, handler)


def sql_driver_connect(connection: Connection, connection_string: str) -> None:
    ansi = False
    if not ansi:
        c_connectString = wchar_pointer(UCS_buf(connection_string))
        f = __lib.SQLDriverConnectW
    else:
        c_connectString = c_char_p(connection_string)
        f = __lib.SQLDriverConnect

    return_code = f(
        connection.handle,
        0,
        c_connectString,
        len(connection_string),
        None,
        0,
        None,
        DriverCompletion.SQL_DRIVER_NOPROMPT.value,
    )
    check_success(return_code, connection)


def sql_disconnect(connection: Connection) -> None:
    check_success(__lib.SQLDisconnect(connection.handle), connection)


def sql_exec_direct(cursor: Cursor, query_string: str) -> None:
    c_query_string = wchar_pointer(UCS_buf(query_string))
    return_code = __lib.SQLExecDirectW(cursor.handle, c_query_string, len(query_string))
    check_success(return_code, cursor)


def sql_row_count(cursor: Cursor) -> int:
    rowcount: SQLLEN = SQLLEN(-1)
    return_code = __lib.SQLRowCount(cursor.handle, byref(rowcount))
    check_success(return_code, cursor)
    return rowcount.value


def sql_num_result_cols(cursor: Cursor) -> int:
    num_cols = SQLSMALLINT(-1)
    check_success(__lib.SQLNumResultCols(cursor.handle, byref(num_cols)), cursor)
    return num_cols.value


@dataclass(frozen=True)
class ColumnDescription:
    name: str
    data_type: DataType  # TODO: data type enum
    size: int
    decimal_digits: int
    is_nullable: bool


def sql_describe_col(cursor: Cursor, column_number: int) -> ColumnDescription:
    buffer_length = 1024
    column_name = create_string_buffer(buffer_length)
    name_length = SQLSMALLINT()
    data_type = SQLSMALLINT()
    column_size = SQLULEN()
    decimal_digits = SQLSMALLINT()
    nullable = SQLSMALLINT()

    return_code = __lib.SQLDescribeCol(
        cursor.handle,
        column_number,
        byref(column_name),
        buffer_length,
        byref(name_length),
        byref(data_type),
        byref(column_size),
        byref(decimal_digits),
        byref(nullable),
    )

    check_success(return_code, cursor)

    column_description = ColumnDescription(
        column_name.value.decode(),
        DataType(data_type.value),
        column_size.value,
        decimal_digits.value,
        bool(nullable.value),
    )

    return column_description


def sql_fetch(cursor: Cursor) -> bool:
    return_code = ReturnCode(__lib.SQLFetch(cursor.handle))
    check_success(return_code, cursor)
    return return_code is not ReturnCode.SQL_NO_DATA


def sql_get_data(
    cursor: Cursor, column_number: int, column_description: ColumnDescription
):
    buffer_size = 4096
    buffer = create_string_buffer(buffer_size)
    foo = SQLLEN()
    return_code = __lib.SQLGetData(
        cursor.handle, column_number, 1, byref(buffer), buffer_size, byref(foo)
    )
    check_success(return_code, cursor)
    # TODO: check return code for SQL_SUCCESS_WITH_INFO (partial data in buffer)
    return buffer.value.decode()


def sql_drivers(
    environment: Environment, ansi: bool = False, include_attributes: bool = False
) -> list[str]:
    buffer_size = 1000
    direction = SqlFetchType.SQL_FETCH_FIRST
    driver_description = create_buffer(buffer_size)
    buffer_length1 = c_short(buffer_size)
    description_length = c_short()
    driver_attributes = create_buffer(buffer_size)
    buffer_length2 = c_short(buffer_size)
    attributes_length = c_short()
    drivers = []
    func = __lib.SQLDrivers if ansi else __lib.SQLDriversW
    while True:
        return_code = func(
            environment.handle,
            direction.value,
            driver_description,
            buffer_length1,
            byref(description_length),
            driver_attributes,
            buffer_length2,
            byref(attributes_length),
        )
        check_success(return_code, environment)
        if ReturnCode(return_code) is ReturnCode.SQL_NO_DATA:
            break

        def decode(buffer) -> str:
            if ansi:
                return buffer.value.decode()
            return UCS_dec(buffer)

        driver = decode(driver_description)
        if include_attributes:
            driver += f" ({decode(driver_attributes)})"
        drivers.append(driver)
        direction = SqlFetchType.SQL_FETCH_NEXT
    return drivers



from typing import List, Dict, Any

from ._connection import Connection
from ._environment import Environment as _Environment
from . import _odbc


apilevel: str = "2.0"
lowercase: bool = False
native_uuid: bool = False
paramstyle: str = "qmark"
pooling: bool = True
threadsafety: int = 1  # TODO: Check threadsafety
version = "0.0.1.dev0"

__environment: _Environment

SQL_CHAR = 1
SQL_WCHAR = -8


def connect(
    connection_string: str,
    autocommit: bool = False,
    ansi: bool = False,
    timeout: int = 0,
    readonly: bool = False,
    attrs_before: Dict[str, Any] = None,
    encoding: str = None,
) -> Connection:

    __ensure_environment_created()

    connection = __environment.connection(
        connection_string, autocommit, ansi, timeout, readonly, attrs_before, encoding
    )

    return connection


def drivers(ansi: bool = False, include_attributes: bool = False) -> List[str]:
    __ensure_environment_created()
    return _odbc.sql_drivers(
        __environment, ansi=ansi, include_attributes=include_attributes
    )


def __ensure_environment_created():
    global __environment
    __environment = _Environment(pooling=pooling)



from __future__ import annotations

import functools
import typing
from dataclasses import dataclass, field

from ._typedef import SQLLEN
from ._row import Row
from . import _odbc
from ._enums import HandleType
from ._handler import Handler
from ._typedef import SQLHSTMT

if typing.TYPE_CHECKING:
    from ._connection import Connection


@dataclass
class Cursor(Handler[SQLHSTMT]):
    connection: Connection
    __rowcount: SQLLEN = field(
        init=False, default_factory=functools.partial(SQLLEN, -1)
    )

    def __post_init__(self):
        _odbc.allocate_statement(self)

    @property
    def rowcount(self) -> int:
        return _odbc.sql_row_count(self)

    @property
    def columncount(self) -> int:
        return _odbc.sql_num_result_cols(self)

    @property
    def handle_type(self) -> HandleType:
        return HandleType.SQL_HANDLE_STMT

    def execute(self, query_string: str) -> Cursor:
        _odbc.sql_exec_direct(self, query_string)
        return self

    def fetchone(self) -> typing.Optional[Row]:
        if not _odbc.sql_fetch(self):
            return None
        row = Row()
        for column_number in range(1, self.columncount + 1):
            # TODO: Other packages call SQLColAttribute here to determine whether int() or long()
            #  is required, but that only applies to Python 2... right? RIGHT???!!
            # _odbc.sql_col_attribute(
            #     cursor,
            #     column_number,
            #     SqlColumnAttrType.SQL_COLUMN_DISPLAY_SIZE
            # )

            # Next, the application retrieves the name, data type, precision,
            # and scale of each result set column with SQLDescribeCol.
            description = _odbc.sql_describe_col(self, column_number)
            value = _odbc.sql_get_data(self, column_number, description)
            setattr(row, description.name, value)
        return row



class Warning(Exception):
    pass


class Error(Exception):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class DataError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class InternalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass



"""This module is more or less a transcription of sqltypes.h"""
import ctypes
import ctypes as c


SIZEOF_LONG_INT = c.sizeof(c.c_long)

# Standard SQL data types.
SQLCHAR = c.c_ubyte
SQLDATE = c.c_ubyte
SQLDECIMAL = c.c_ubyte
SQLDOUBLE = c.c_double
SQLFLOAT = c.c_double
SQLINTEGER = c.c_int
SQLUINTEGER = c.c_uint
SQLLEN = c.c_long
SQLULEN = c.c_ulong
SQLSETPOSIROW = c.c_ulong
SQLROWCOUNT = SQLULEN
SQLROWSETSIZE = SQLULEN
SQLTRANSID = SQLULEN
SQLROWOFFSET = SQLULEN
SQLNUMERIC = c.c_byte
SQLPOINTER = c.c_void_p
SQLREAL = c.c_float
SQLSMALLINT = c.c_short
SLQUSMALLINT = c.c_ushort
SQLTIME = c.c_ubyte
SQLTIMESTAMP = c.c_ubyte
SQLVARCHAR = c.c_ubyte
SQLRETURN = SQLSMALLINT
SQLHANDLE = c.c_void_p
SQLHENV = SQLHANDLE
SQLHDBC = SQLHANDLE
SQLHSTMT = SQLHANDLE
SQLHDESC = SQLHANDLE

# More basic data types to augment what windows.h provides
UCHAR = ctypes.c_ubyte
SCHAR = c.c_char
SQLSCHAR = SCHAR
if SIZEOF_LONG_INT == 4:
    SDWORD = c.c_long
    UDWORD = c.c_ulong
else:
    SDWORD = c.c_int
    UDWORD = c.c_uint
SWORD = c.c_short
UWORD = c.c_ushort
UINT = c.c_uint
SLONG = c.c_long
SSHORT = c.c_short
ULONG = c.c_ulong
USHORT = c.c_ushort
SDOUBLE = c.c_double
LDOUBLE = c.c_double
SFLOAT = c.c_float
PTR = c.c_void_p
RETCODE = c.c_short
SQLHWND = c.c_void_p
SQLUSMALLINT = c.c_ushort



from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Generic

from ._custom_types import TSqlHandle
from ._typedef import SQLHANDLE
from ._enums import HandleType
from . import _odbc


@dataclass  # type: ignore
class Handler(Generic[TSqlHandle]):
    """Python object which references a SQLHANDLE."""

    handle: TSqlHandle = field(init=False, default_factory=SQLHANDLE)
    _closed: bool = field(init=False, default=False)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    @abstractmethod
    def handle_type(self) -> HandleType:
        ...

    def close(self) -> None:
        if self._closed:
            return
        _odbc.sql_free_handle(self)
        self._closed = True



from dataclasses import dataclass

from ._errors import ProgrammingError
from . import _odbc
from ._cursor import Cursor
from ._handler import Handler
from ._enums import HandleType
from ._typedef import SQLHDBC


@dataclass
class Connection(Handler[SQLHDBC]):
    """Connection objects manage connections to the database.
    Each manages a single ODBC HDBC.
    """

    def cursor(self) -> Cursor:
        cur = Cursor(self)
        return cur

    @property
    def handle_type(self) -> HandleType:
        return HandleType.SQL_HANDLE_DBC

    def close(self) -> None:
        _odbc.sql_disconnect(self)
        super().close()

    def setencoding(self, *args, **kwargs):
        raise NotImplementedError

    def setdecoding(self, sqltype, encoding, ctype) -> None:
        """
        setdecoding(sqltype, encoding=None, ctype=None) --> None

        Configures how text of type `ctype` (SQL_CHAR or SQL_WCHAR) is decoded
        when read from the database.

        When reading, the database will assign one of the sqltypes to text columns.
        pyodbc uses this lookup the decoding information set by this function.
        sqltype: pyodbc.SQL_CHAR or pyodbc.SQL_WCHAR

        encoding: A registered Python encoding such as "utf-8".

        ctype: The C data type should be requested.  Set this to SQL_CHAR for
          single-byte encodings like UTF-8 and to SQL_WCHAR for two-byte encodings
          like UTF-16.
        """
        raise NotImplementedError


def connection_check(cnxn: Connection) -> bool:
    return isinstance(cnxn, Connection)


def connection_validate(cnxn: Connection) -> Connection:
    if not cnxn or not connection_check(cnxn):
        raise TypeError("Connection object required")

    if not cnxn.handle.value:
        raise ProgrammingError("Attempt to use a closed connection.")

    return cnxn



from __future__ import annotations

from dataclasses import dataclass

from . import _odbc
from ._enums import HandleType, OdbcVersion
from ._handler import Handler
from ._connection import Connection
from ._typedef import SQLHENV


@dataclass
class Environment(Handler[SQLHENV]):
    pooling: bool

    def __post_init__(self):
        _odbc.allocate_environment(self)
        _odbc.set_environment_odbc_version(self, OdbcVersion.SQL_OV_ODBC3_80)

    def connection(
        self,
        connection_string: str,
        autocommit: bool = False,
        ansi: bool = False,
        timeout: int = 0,
        readonly: bool = False,
        attrs_before: dict = None,
        encoding: str = None,
    ) -> Connection:
        connection = Connection()
        _odbc.allocate_connection(self, connection)
        _odbc.sql_driver_connect(connection, connection_string)
        return connection

    @property
    def handle_type(self) -> HandleType:
        return HandleType.SQL_HANDLE_ENV



from enum import Enum


class OdbcVersion(Enum):
    SQL_OV_ODBC2 = 2
    SQL_OV_ODBC3 = 3
    SQL_OV_ODBC3_80 = 380


class DataType(Enum):
    """https://docs.microsoft.com/en-us/sql/odbc/reference/appendixes/sql-data-types"""

    SQL_TINYINT = -6
    SQL_BIT = -7
    # SQL_WCHAR = -8
    SQL_WVARCHAR = -9
    # SQL_WLONGVARCHAR = -10
    SQL_CHAR = 1
    SQL_INTEGER = 4
    SQL_VARCHAR = 12
    SQL_TYPE_TIMESTAMP = 93


class EnvironmentAttributeType(Enum):
    """https://docs.microsoft.com/en-us/sql/odbc/reference/syntax/sqlsetenvattr-function"""

    SQL_ATTR_ODBC_VERSION = 200
    SQL_ATTR_CONNECTION_POOLING = 201
    SQL_ATTR_CP_MATCH = 202
    SQL_ATTR_OUTPUT_NTS = 10001


class DriverCompletion(Enum):
    SQL_DRIVER_NOPROMPT = 0
    SQL_DRIVER_COMPLETE = 1
    SQL_DRIVER_PROMPT = 2
    SQL_DRIVER_COMPLETE_REQUIRED = 3


class ReturnCode(Enum):
    SQL_INVALID_HANDLE = -2
    SQL_ERROR = -1
    SQL_SUCCESS = 0
    SQL_SUCCESS_WITH_INFO = 1
    SQL_NO_DATA = 100
    SQL_STILL_EXECUTING = 2


class HandleType(Enum):
    SQL_NULL_HANDLE = 0
    SQL_HANDLE_ENV = 1
    SQL_HANDLE_DBC = 2
    SQL_HANDLE_STMT = 3
    SQL_HANDLE_DESC = 4
    # SQL_HANDLE_DBC_INFO_TOKEN = ?


class SqlColumnAttrType(Enum):
    SQL_COLUMN_NAME = 1
    SQL_COLUMN_TYPE = 2
    SQL_COLUMN_LENGTH = 3
    SQL_COLUMN_PRECISION = 4
    SQL_COLUMN_SCALE = 5
    SQL_COLUMN_DISPLAY_SIZE = 6
    SQL_COLUMN_NULLABLE = 7
    SQL_COLUMN_UNSIGNED = 8
    SQL_COLUMN_MONEY = 9
    SQL_COLUMN_UPDATABLE = 10
    SQL_COLUMN_AUTO_INCREMENT = 11
    SQL_COLUMN_CASE_SENSITIVE = 12
    SQL_COLUMN_SEARCHABLE = 13
    SQL_COLUMN_TYPE_NAME = 14
    SQL_COLUMN_TABLE_NAME = 15
    SQL_COLUMN_OWNER_NAME = 16
    SQL_COLUMN_QUALIFIER_NAME = 17
    SQL_COLUMN_LABEL = 18


class SqlFetchType(Enum):
    SQL_FETCH_NEXT = 1
    SQL_FETCH_FIRST = 2
    SQL_FETCH_LAST = 3
    SQL_FETCH_PRIOR = 4
    SQL_FETCH_ABSOLUTE = 5
    SQL_FETCH_RELATIVE = 6
